/* Copyright (C) 2009 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _DATA_H_
#define _DATA_H_

#include "params.h"
#include "fbvar.h"

#include <fstream>

#include <gsl/gsl_rng.h>
#include <string>
#include <vector>

using namespace std;

class Data
{

 public:
  int _nS;
  int _nR;
  float _xmin;
  float _xmax;
  int* _obs_status; //missing 0; OK 1; censored 2 or 3
  int _lseq;
  string* _pos;   // length lseq
  float* _x;  // length lseq
  float* _rx; // length lseq
  float* _seqs; // length lseq
  string _datafilename;  

  float _CIlevel;
  float* _expectS;
  float* _postpoutlier;
  int* _mostlikelyS;
  int* _mostlikelyR;
  float** _posteriorR;
  int* _minCI;
  int* _maxCI;
  int* _viterbiR;
  int* _viterbiM;
  int* _viterbiS;
  float* _posteriorM1u;
  float* _posteriorM1d;
  float* _posteriorM2;
  float* _posteriorM3;

 public:
  Data(string datafilename); 
  ~Data();
  void allocalgovar(int nS, int nR, float xmin, float xmax, float absrxmax, float seqsmax); 
  vector<float> range();
  double forward(Params& par, FbVar& fbv, int imin, int imax, int irecmin, int irecmax); // return loglik
  void backward(Params& par, FbVar& fbv, int imin, int imax, int irecmin, int irecmax);
  double FB(Params& par, FbVar& fbv, int estep_l, int estep_ov);
  void CIpos(int t, float** pvect, Params& par);  

  void viterbi(Params& par, FbVar& fbv, int imin, int imax, int irecmin, int irecmax);
  void VIT(Params& par, FbVar& fbv, int estep_l, int estep_ov);

  inline int lseq();
  void printPath(Params& par, ofstream& out);
  void count_sum_sd(Params& par,int t, float** pvect, float* pvectobs);
  void count_t(Params& par, int t);
  int nxinfmu(float mu);
  float absrxmax();
};

inline int Data::lseq() 
{
  return(_lseq);
}

#endif
