/* Copyright (C) 2009 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream>
#include <fstream>
#include <cstdlib>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>

#include "data.h"
#include "params.h"
#include "fbvar.h"

using namespace std;


int 
main(int argc, char** argv) 
{
  string datafilename1;
  string datafilename2;
  int nRstates = 1;
  int nSstates = 100;
  float xrangefracX = 0.9;
  float xrangefracS = 1.1;
  float absrxtrunc = 200;
  int itermax = 100;
  string oprefix("out");
  double frac_cat_mu=0.13;
  string iparamstr("none");
  time_t seed = time(NULL);
  double epsmax = 0.01;
  float seqsmax = 1.9;
  int outlier_type = 2;
  int sigma_type = 2;
  int rho_type = 2;
  int nexplo = 1;
  int niterexplo = 100;
  float epsexplo = 0.1;
  char randompar = 'F';
  bool etaflat = false;

  vector<string> datafilev;
  if (argc>=3)
    {
      int iarg=1;
      string keyword;
      while (iarg < argc) {
	keyword.assign(argv[iarg]);
	if (keyword.compare("-data") == 0) {
	  string word(argv[++iarg]);
	  while ((iarg < argc) && (word[0] != '-')) {
	    datafilev.push_back(word);
	    iarg++;
	    if (iarg < argc) {
	      word.assign(argv[iarg]);
	    }
	  }
	} else if (keyword.compare("-nR") == 0) {
	  nRstates = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-nS") == 0) {
	  nSstates = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-itermax") == 0) {
	  itermax = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-Xrange") == 0) {
	  xrangefracX = atof(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-Srange") == 0) {
	  xrangefracS = atof(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-rtrunc") == 0) {
	  absrxtrunc = atof(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-eta") == 0) {
	  string word(argv[++iarg]);
	  if (word.compare("flat") == 0) {
	    etaflat = true;
	  } else if (word.compare("kernel") != 0) {
	    cerr << "Aborted: unknown value for arg -eta : '" << word << "'" << endl;
	    exit(1);
	  }
	  iarg++;
	} else if (keyword.compare("-oprefix") == 0) {
	  oprefix.assign(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-fraccat") == 0) {
	  frac_cat_mu = atof(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-iparam") == 0) {
	  iparamstr.assign(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-randompar") == 0) {
	  randompar = argv[++iarg][0];
	  iarg++;
	} else if (keyword.compare("-seed") == 0) {
	  seed = (time_t)atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-eps") == 0) {
	  epsmax = atof(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-rhotype") == 0) {
	  rho_type = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-sigmatype") == 0) {
	  sigma_type = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-outliertype") == 0) {
	  outlier_type = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-nexplo") == 0) {
	  nexplo = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-niterexplo") == 0) {
	  niterexplo = atoi(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-epsexplo") == 0) {
	  epsexplo = atof(argv[++iarg]);
	  iarg++;
	} else if (keyword.compare("-seqsmax") == 0) {
	  seqsmax = atof(argv[++iarg]);
	  iarg++;
	} else {
	  cerr << "Aborted: unexpected word in command line " << keyword << endl;
	  exit(1);
	}
      }
    } else {
      cout << "Usage: " << argv[0] << endl
	   << " -data <file1> [<file2> ...]" << endl
	   << " [-nR <nR>] (default 1)" << endl
	   << " [-nS <nS>] (default 100)" << endl
	   << " [-Xrange <xr>] (default 0.9)" << endl
	   << " [-Srange <sr>] (default 1.1)" << endl
	   << " [-rtrunc <rm>] (default 200)" << endl 
	   << " [-itermax <im>] (default 100)" << endl
	   << " [-oprefix <op>] (default out)" << endl 
	   << " [-fraccat <fc>] (default 0.13)" << endl 
	   << " [-iparam <filename>] (default none)" << endl
	   << " [-randompar <T or F>] (default F)" << endl
	   << " [-seed <s>] (default time(NULL))" << endl
	   << " [-eta <kernel or flat>] (default kernel)" << endl
	   << " [-rhotype <rhotype>] (default 2)" << endl
	   << " [-sigmatype <sigmatype>] (default 2)" << endl
	   << " [-outliertype <outliertype>] (default 2)" << endl
	   << " [-eps <epsmax>] (default 0.01)" << endl
	   << " [-nexplo <nexplo>] (default 10)" << endl
	   << " [-niterexplo <niterexplo>] (default 100)" << endl
	   << " [-epsexplo <epsexplo>] (default 0.1)" << endl
	   << " [-seqsmax <seqsmax>] (default 1.9)" << endl;
      exit(1);
    }

  if (xrangefracX<0 || xrangefracX>1) {
    cerr << "the value of the parameter xrangefracX should be between 0 and 1" << endl;
    exit(1);
  } 
  if (xrangefracS<1) {
    cerr << "the value of the parameter xrangefracS should be above 1" << endl;
    exit(1);
  } 

  if (datafilev.size()<1) {
    cerr << "You should specify at least one data file" << endl;
    exit(1);
  }

  cout << "datafilev ";
  for (unsigned int ifn=0; ifn<datafilev.size(); ifn++) {
    cout << " " << datafilev[ifn];
  }
  cout << endl;
  cout  << "nRstates " << nRstates << endl
	<< "nSstates " << nSstates << endl
	<< "xrangefracX " << xrangefracX << endl
	<< "xrangefracS " << xrangefracS << endl
	<< "absrxtrunc " << absrxtrunc << endl
	<< "itermax " << itermax << endl
	<< "oprefix " << oprefix << endl
	<< "fraccat " << frac_cat_mu << endl
	<< "iparam " << iparamstr << endl
	<< "randompar " << randompar << endl
	<< "eps " << epsmax << endl
	<< "seqsmax " << seqsmax << endl
	<< "nexplo " << nexplo << endl
	<< "niterexplo " << niterexplo << endl
	<< "epsexplo " <<  epsexplo << endl
	<< endl;


  cout << "seed " << seed << endl;
  
  int ndata = (int)datafilev.size();
  Data** datav = new Data* [ndata];
  for (int id=0; id<ndata; id++) {
    datav[id] = new Data(datafilev[id]);
  }


  float rxmax = datav[0]->absrxmax();
  for (int id=1; id<ndata; id++) {
    if (datav[id]->absrxmax()>rxmax) {
      rxmax = datav[id]->absrxmax();
    }
  }
  int ncatrx = 100*(int)(rxmax+1);
  if (randompar=='F') {
    nexplo=1;
  }
  Params** vparams = new Params* [nexplo];
  if (iparamstr.compare("none")!=0) {
    for (int ipar=0; ipar<nexplo; ipar++) {
      ifstream iparams (iparamstr.c_str(), ios::in);
      vparams[ipar] = new Params(iparams);
      iparams.close();
      vparams[ipar]->alloclogit(ncatrx);
    } 
    if (randompar=='T') {
      gsl_rng * r = gsl_rng_alloc (gsl_rng_taus);
      gsl_rng_set(r,seed);

      for (int ipar=0; ipar<nexplo; ipar++) {
	vparams[ipar]->initrandom(r);
	vparams[ipar]->alloclogit(ncatrx);
      }
      gsl_rng_free(r);      
    }
  } else {
    vector<float> range = datav[0]->range();
    for (int id=1; id<ndata; id++) {
      vector<float> rangei = datav[id]->range();    
      if (rangei[0]<range[0]) {
	range[0]=rangei[0];
      }
      if (rangei[1]>range[1]) {
	range[1]=rangei[1];
      }
    }  
    float xmin = range[0]+(range[1]-range[0])*(1-xrangefracX)/2;
    float xmax = range[1]-(range[1]-range[0])*(1-xrangefracX)/2;
    float smin = range[0]-(range[1]-range[0])*(xrangefracS-1)/2;
    float smax = range[1]+(range[1]-range[0])*(xrangefracS-1)/2;
    if (randompar=='F') {
      vparams[0] = new Params(nSstates, nRstates, xmin, xmax, smin, smax);
      vparams[0]->definecatmu(frac_cat_mu, datav, ndata);
      vparams[0]->definefstatkdens(datav, ndata, etaflat);
      vparams[0]->initdeterministic(sigma_type,rho_type,outlier_type);
      vparams[0]->alloclogit(ncatrx);      
    } else if (randompar=='T') {
      gsl_rng * r = gsl_rng_alloc (gsl_rng_taus);
      gsl_rng_set(r,seed);

      vparams[0] = new Params(nSstates, nRstates, xmin, xmax, smin, smax);
      vparams[0]->definefstatkdens(datav, ndata, etaflat);
      vparams[0]->definecatmu(frac_cat_mu, datav, ndata);
      vparams[0]->initrandom(sigma_type,rho_type,outlier_type, r);
      vparams[0]->alloclogit(ncatrx);

      for (int ipar=1; ipar<nexplo; ipar++) {
	vparams[ipar] = new Params(nSstates, nRstates, xmin, xmax, smin, smax);
	vparams[ipar]->copycatmufstat(*vparams[0]);
	vparams[ipar]->initrandom(sigma_type,rho_type,outlier_type, r);
	vparams[ipar]->alloclogit(ncatrx);
      }
      gsl_rng_free(r);      
    }
  }

  for (int id=0; id<ndata; id++) {
    datav[id]->allocalgovar(vparams[0]->_nS, vparams[0]->_nR, vparams[0]->_xmin, vparams[0]->_xmax, absrxtrunc, seqsmax);
  }

  FbVar fbv(*(vparams[0]), 10000); 

  int bestparams = 0;
  if (nexplo > 1) {
    double bestlogl = 0;
    string straceexplo(oprefix);
    straceexplo+="_params.explotrace";
    ofstream otraceexplo (straceexplo.c_str(), ios::out);
    
    for (int ipar=0; ipar<nexplo; ipar++) {
      double loglikexplo=0;
      double newloglikexplo;
      double eps;
      bool keepgoingexplo=true;
      for (int iter=0; keepgoingexplo; iter++) {
	vparams[ipar]->print(otraceexplo);
	//cout << "time " << time(NULL) << endl;
	vparams[ipar]->reset_counts();
	newloglikexplo=0;
	for (int id=0; id<ndata; id++) {
	  newloglikexplo+=datav[id]->FB(*(vparams[ipar]), fbv, 10000, 500);
	}
	eps = newloglikexplo-loglikexplo;
	cout << " explo " << ipar << " iter " << iter << " loglik " << newloglikexplo;
	otraceexplo << "# explo " << ipar << " iter " << iter << " loglik " << newloglikexplo;
	if (iter>=1) {
	  cout << " eps " << eps;
	  otraceexplo << " eps " << eps;
	}
	cout << endl;
	otraceexplo << endl;
	loglikexplo = newloglikexplo;
	if (iter>=1 && eps<epsexplo) {keepgoingexplo=false;}
	if (iter>=niterexplo) {keepgoingexplo=false;}
	if (keepgoingexplo) {
	  vparams[ipar]->m_step(iter);
	}
      }
      cout << " final explo " << ipar << " logl " << loglikexplo << endl;
      otraceexplo << "# final explo " << ipar << " logl " << loglikexplo << endl;
      if (ipar==0 || loglikexplo>bestlogl) {
	bestlogl=loglikexplo;
	bestparams=ipar;
      }
    }
    otraceexplo.close();
  }

  Params* params = vparams[bestparams];
  string strace(oprefix);
  strace+="_params.trace";
  ofstream otrace (strace.c_str(), ios::out);

  double loglik=0;
  double newloglik;
  double eps;
  bool keepgoing=true;
  for (int iter=0; keepgoing; iter++) {
    params->print(otrace);
    //cout << "time " << time(NULL) << endl;
    params->reset_counts();
    newloglik=0;
    for (int id=0; id<ndata; id++) {
      newloglik+=datav[id]->FB(*params, fbv, 10000, 500);
    }
    eps = newloglik-loglik;
    cout << " iter " << iter << " loglik " << newloglik;
    otrace << "# iter " << iter << " loglik " << newloglik;
    if (iter>=1) {
      cout << " eps " << eps;
      otrace << " eps " << eps;
    }
    cout << endl;
    otrace << endl;
    loglik = newloglik;
    if (iter>=1 && eps<epsmax) {keepgoing=false;}
    if (iter>=itermax) {keepgoing=false;}
    if (keepgoing) {
      params->m_step(iter);
    }
  }

  otrace.close();

  string soparams(oprefix);
  soparams+=".oparams";
  ofstream oparams (soparams.c_str(), ios::out);
  params->print(oparams);
  oparams.close();

  for (int id=0; id<ndata; id++) {
    datav[id]->VIT(*params, fbv, 10000, 500);
    string sopath(oprefix);
    sopath+=datav[id]->_datafilename;
    sopath+=".path";
    ofstream opath (sopath.c_str(), ios::out);
    datav[id]->printPath(*params, opath);
    opath.close();
  }
  return 1;
}

